"""Schedule plugin

Provides duty schedule / duty roster functions
"""
from datetime import datetime, timedelta, timezone
from http.client import NO_CONTENT, CREATED
from flask import Blueprint, request, jsonify
from werkzeug.exceptions import BadRequest, NotFound
from backports.datetime_fromisoformat import MonkeyPatch

from flaschengeist.plugins import Plugin
from flaschengeist.decorator import login_required

from . import event_controller, permissions
from . import models

MonkeyPatch.patch_fromisoformat()
schedule_bp = Blueprint("schedule", __name__, url_prefix="/schedule")


class SchedulePlugin(Plugin):
    models = models

    def __init__(self, config):
        super().__init__(
            blueprint=schedule_bp,
            permissions=permissions.permissions,
        )


@schedule_bp.route("/event-types", methods=["GET"])
@login_required()
def get_event_types(current_session):
    """Get all EventTypes

    Route: ``/event-types`` | Method: ``GET``

    Args:
        current_session: Session sent with Authorization Header

    Returns:
        JSON encoded list of EventTypes HTTP-error
    """
    types = event_controller.get_event_types()
    return jsonify(types)


@schedule_bp.route("/event-types", methods=["POST"])
@login_required(permission=permissions.EVENT_TYPE)
def new_event_type(current_session):
    """Create a new EventType

    Route: ``/event-types`` | Method: ``POST``

    POST-data: ``{name: string}``

    Args:
        current_session: Session sent with Authorization Header

    Returns:
        HTTP-Created or HTTP-error
    """
    data = request.get_json()
    if "name" not in data:
        raise BadRequest
    event_controller.create_event_type(data["name"])
    return "", CREATED


@schedule_bp.route("/event-types/<name>", methods=["PUT", "DELETE"])
@login_required(permission=permissions.EVENT_TYPE)
def modify_event_type(name, current_session):
    """Rename or delete an event type

    Route: ``/event-types/<name>`` | Method: ``PUT`` or ``DELETE``

    POST-data: (if renaming) ``{name: string}``

    Args:
        name: Name identifying the EventType
        current_session: Session sent with Authorization Header

    Returns:
        HTTP-NoContent or HTTP-error
    """
    if request.method == "DELETE":
        event_controller.delete_event_type(name)
    else:
        data = request.get_json()
        if "name" not in data:
            raise BadRequest("Parameter missing in data")
        event_controller.rename_event_type(name, data["name"])
    return "", NO_CONTENT


@schedule_bp.route("/job-types", methods=["GET"])
@login_required()
def get_job_types(current_session):
    """Get all JobTypes

    Route: ``/job-types`` | Method: ``GET``

    Args:
        current_session: Session sent with Authorization Header

    Returns:
        JSON encoded list of JobType HTTP-error
    """
    types = event_controller.get_job_types()
    return jsonify(types)


@schedule_bp.route("/job-types", methods=["POST"])
@login_required(permission=permissions.JOB_TYPE)
def new_job_type(current_session):
    """Create a new JobType

    Route: ``/job-types`` | Method: ``POST``

    POST-data: ``{name: string}``

    Args:
        current_session: Session sent with Authorization Header

    Returns:
        HTTP-Created or HTTP-error
    """
    data = request.get_json()
    if "name" not in data:
        raise BadRequest
    event_controller.create_job_type(data["name"])
    return "", CREATED


@schedule_bp.route("/job-types/<name>", methods=["PUT", "DELETE"])
@login_required(permission=permissions.JOB_TYPE)
def modify_job_type(name, current_session):
    """Rename or delete a JobType

    Route: ``/job-types/<name>`` | Method: ``PUT`` or ``DELETE``

    POST-data: (if renaming) ``{name: string}``

    Args:
        name: Name identifying the JobType
        current_session: Session sent with Authorization Header

    Returns:
        HTTP-NoContent or HTTP-error
    """
    if request.method == "DELETE":
        event_controller.delete_name_type(name)
    else:
        data = request.get_json()
        if "name" not in data:
            raise BadRequest("Parameter missing in data")
        event_controller.rename_name_type(name, data["name"])
    return "", NO_CONTENT


@schedule_bp.route("/events/<int:event_id>", methods=["GET"])
@login_required()
def get_event(event_id, current_session):
    """Get event by id

    Route: ``/events/<event_id>`` | Method: ``GET``

    Args:
        event_id: ID identifying the event
        current_session: Session sent with Authorization Header

    Returns:
        JSON encoded event object
    """
    event = event_controller.get_event(event_id)
    return jsonify(event)


@schedule_bp.route("/events", methods=["GET"])
@schedule_bp.route("/events/<int:year>/<int:month>", methods=["GET"])
@schedule_bp.route("/events/<int:year>/<int:month>/<int:day>", methods=["GET"])
@login_required()
def get_events(current_session, year=datetime.now().year, month=datetime.now().month, day=None):
    """Get Event objects for specified date (or month or year),
     if nothing set then events for current month are returned

    Route: ``/events[/<year>/<month>[/<int:day>]]`` | Method: ``GET``

    Args:
        year (int, optional): year to query, defaults to current year
        month (int, optional): month to query (if set), defaults to current month
        day (int, optional): day to query events for (if set)
        current_session: Session sent with Authorization Header

    Returns:
        JSON encoded list containing events found or HTTP-error
    """
    try:
        begin = datetime(year=year, month=month, day=1)
        if day:
            begin += timedelta(days=day - 1)
            end = begin + timedelta(days=1)
        else:
            if month == 12:
                end = datetime(year=year + 1, month=1, day=1)
            else:
                end = datetime(year=year, month=month + 1, day=1)

        events = event_controller.get_events(begin, end)
        return jsonify(events)
    except ValueError:
        raise BadRequest("Invalid date given")


def _event_slot_from_data(data):
    try:
        start = datetime.fromisoformat(data["start"]).replace(tzinfo=timezone.utc)
        end = datetime.fromisoformat(data["end"]).replace(tzinfo=timezone.utc) if "end" in data else None
        # jobs = ...
    except (NotFound, KeyError, ValueError):
        raise BadRequest("Missing POST parameter")
    return {"start": start, "end": end}


@schedule_bp.route("/events", methods=["POST"])
@login_required(permission=permissions.CREATE)
def create_event(current_session):
    """Create an new event

    Route: ``/events`` | Method: ``POST``

    POST-data: See interfaces for Event, can already contain slots

    Args:
        current_session: Session sent with Authorization Header

    Returns:
        JSON encoded Event object or HTTP-error
    """
    data = request.get_json()
    try:
        start = datetime.fromisoformat(data["start"]).replace(tzinfo=timezone.utc)
        event_type = event_controller.get_event_type(data["type"])
    except (NotFound, KeyError, ValueError):
        raise BadRequest("Missing POST parameter")

    data = request.get_json()
    event = event_controller.create_event(
        start=start, event_type=event_type, description=data["description"] if "description" in data else None
    )
    if "slots" in data:
        for slot in data["slots"]:
            event_controller.add_event_slot(event, **_event_slot_from_data(slot))
    return jsonify(event)


@schedule_bp.route("/events/<int:event_id>", methods=["PUT"])
@login_required(permission=permissions.EDIT)
def modify_event(event_id, current_session):
    """Modify an event

    Route: ``/events/<event_id>`` | Method: ``PUT``

    POST-data: See interfaces for Event, can already contain slots

    Args:
        event_id: Identifier of the event
        current_session: Session sent with Authorization Header

    Returns:
        JSON encoded Event object or HTTP-error
    """
    event = event_controller.get_event(event_id)
    data = request.get_json()
    if "start" in data:
        event.start = datetime.fromisoformat(data["start"]).replace(tzinfo=timezone.utc)
    if "description" in data:
        event.description = data["description"]
    if "type" in data:
        event_type = event_controller.get_event_type(data["type"])
        event.type = event_type
    event_controller.update()
    return jsonify(event)


@schedule_bp.route("/events/<int:event_id>", methods=["DELETE"])
@login_required(permission=permissions.DELETE)
def delete_event(event_id, current_session):
    """Delete an event

    Route: ``/events/<event_id>`` | Method: ``DELETE``

    Args:
        event_id: Identifier of the event
        current_session: Session sent with Authorization Header

    Returns:
        HTTP-NoContent or HTTP-error
    """
    event_controller.delete_event(event_id)
    return "", NO_CONTENT


@schedule_bp.route("/events/<int:event_id>/slots", methods=["POST"])
@login_required(permission=permissions.EDIT)
def add_event_slot(event_id, current_session):
    """Add an new EventSlot to an Event

    Route: ``/events/<event_id>/slots`` | Method: ``POST``

    POST-data: See TS interface for EventSlot

    Args:
        event_id: Identifier of the event
        current_session: Session sent with Authorization Header

    Returns:
        JSON encoded Event object or HTTP-error
    """
    event = event_controller.get_event(event_id)
    data = request.get_json()
    if not data:
        raise BadRequest("Missing POST parameters")

    event_controller.add_event_slot(event, **_event_slot_from_data(data))
    return jsonify(event)


@schedule_bp.route("/events/<int:event_id>/slots/<int:slot_id>", methods=["PUT"])
@login_required(permission=permissions.EDIT)
def update_event_slot(event_id, slot_id, current_session):
    """Update an EventSlot

    Route: ``/events/<event_id>/slots/<slot_id>`` | Method: ``PUT``

    POST-data: See TS interface for EventSlot

    Args:
        event_id: Identifier of the event
        slot_id: Identifier of the slot
        current_session: Session sent with Authorization Header

    Returns:
        JSON encoded Event object or HTTP-error
    """
    event = event_controller.get_event(event_id)
    slot = event_controller.get_event_slot(slot_id)
    if slot not in event.slots:
        raise NotFound

    data = _event_slot_from_data(request.get_json())
    slot.start = data["start"]
    if "end" in data:
        slot.end = data["end"]
    event_controller.update()
    return jsonify(event)


@schedule_bp.route("/events/<int:event_id>/slots/<int:slot_id>", methods=["DELETE"])
@login_required(permission=permissions.EDIT)
def delete_event_slot(event_id, slot_id, current_session):
    """Delete an EventSlot

   Route: ``/events/<event_id>/slots/<slot_id>`` | Method: ``DELETE``

   Args:
       event_id: Identifier of the event
       slot_id: Identifier of the slot
       current_session: Session sent with Authorization Header

   Returns:
       JSON encoded Event object or HTTP-error
   """
    event = event_controller.get_event(event_id)
    event_controller.remove_event_slot(event, slot_id)
    return jsonify(event)


# TODO: JobSlot, Job!
# TODO: JobTransfer
