from datetime import datetime
from typing import Optional

from flaschengeist.models import ModelSerializeMixin, UtcDateTime
from flaschengeist.models.user import User
from flaschengeist.database import db

#########
# Types #
#########


class EventType(db.Model, ModelSerializeMixin):
    __tablename__ = "event_type"
    id_: int = db.Column("id", db.Integer, primary_key=True)
    name: str = db.Column(db.String(30), nullable=False, unique=True)


class JobType(db.Model, ModelSerializeMixin):
    __tablename__ = "job_type"
    id_ = db.Column("id", db.Integer, primary_key=True)
    name: str = db.Column(db.String(30), nullable=False, unique=True)


########
# Jobs #
########


class Job(db.Model, ModelSerializeMixin):
    __tablename__ = "job"
    userid: str = None
    value: float = db.Column(db.Numeric(precision=3, scale=2), nullable=False)

    id_ = db.Column("id", db.Integer, primary_key=True)
    _user_id = db.Column("user_id", db.Integer, db.ForeignKey("user.id"), nullable=False)
    _slot_id = db.Column("slot_id", db.Integer, db.ForeignKey("job_slot.id"), nullable=False)

    user_: User = db.relationship("User")
    slot_ = db.relationship("JobSlot")

    @property
    def userid(self):
        return self._user.userid


class JobSlot(db.Model, ModelSerializeMixin):
    __tablename__ = "job_slot"
    _type_id = db.Column("type_id", db.Integer, db.ForeignKey("job_type.id"))
    _event_slot_id = db.Column("event_slot_id", db.Integer, db.ForeignKey("event_slot.id"))

    id: int = db.Column(db.Integer, primary_key=True)
    type: JobType = db.relationship("JobType")
    users: [Job] = db.relationship("Job", back_populates="slot_")
    required_jobs: float = db.Column(db.Numeric(precision=4, scale=2))

    event_slot_ = db.relationship("EventSlot", back_populates="jobs")


##########
# Events #
##########


class EventSlot(db.Model, ModelSerializeMixin):
    """Model for an EventSlot"""

    __tablename__ = "event_slot"
    _event_id = db.Column("event_id", db.Integer, db.ForeignKey("event.id"), nullable=False)

    id: int = db.Column(db.Integer, primary_key=True)
    start: datetime = db.Column(UtcDateTime)
    end: Optional[datetime] = db.Column(UtcDateTime)
    jobs: [JobSlot] = db.relationship("JobSlot", back_populates="event_slot_")

    event_ = db.relationship("Event", back_populates="slots")


class Event(db.Model, ModelSerializeMixin):
    """Model for an Event"""

    __tablename__ = "event"
    _type_id = db.Column("type_id", db.Integer, db.ForeignKey("event_type.id", ondelete="CASCADE"), nullable=False)

    id: int = db.Column(db.Integer, primary_key=True)
    start: datetime = db.Column(UtcDateTime, nullable=False)
    description: Optional[str] = db.Column(db.String(240))
    type: EventType = db.relationship("EventType")
    slots: [EventSlot] = db.relationship(
        "EventSlot", back_populates="event_", cascade="all,delete,delete-orphan", order_by="EventSlot.start"
    )
